# Copyright (C) 2020 jumanjiman (Paul Morgan) <jumanjiman@gmail.com>
#
# This file is part of foal.
#
# foal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# foal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with foal.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""Library to merge acronyms from YAML files."""
import glob
import logging
import os
import sys

from identify import identify
from ruamel.yaml import YAML

# This filename takes precedence over all others.
ROOT_GLOBAL = "global.yaml"


class Acronyms:
    """
    Read acronyms from all YAML files in a directory tree.

    From the user perspective, "first acronym definition wins".
    """

    def __init__(self, prefer_files=None, exclude_files=None):
        """Return a new Acronyms object to use for lookups."""
        self.search_list = self._search_list(prefer_files, exclude_files)
        self.merged_acronyms = dict(sorted(self._merged_acronyms().items()))

    def get(self, acronyms=None, ignorecase=False):
        """Get definitions for acronynms (or all definitions if acronyms is empty)."""
        result = self.merged_acronyms
        if not acronyms:
            return result

        if ignorecase:
            return {
                k: v for k, v in result.items()
                if k.lower() in [a.lower() for a in acronyms]
            }

        return {k: v for k, v in result.items() if k in acronyms}

    @staticmethod
    def acronyms_from(path):
        """Load definitions from a file. Return a dictionary."""
        try:
            with open(path, "r") as stream:
                content = YAML(typ="safe").load(stream)
        except FileNotFoundError as err:
            # File has disappeared or was incorrectly put in exclude_flies or prefer_files.
            logging.error("%s: %s", err.strerror, err.filename)
            sys.exit(1)

        acronyms = content.get("acronyms", {})
        for _, val in acronyms.items():
            val["source"] = path  # Show where acronym came from.
        return acronyms

    @staticmethod
    def base_filename(path):
        """Strip directory and file extension from path."""
        basename = os.path.basename(path)
        filename = os.path.splitext(basename)[0]
        return filename

    @staticmethod
    def is_yaml(path):
        """Return True if path is a YAML file."""
        try:
            result = "yaml" in identify.tags_from_path(path)
        except OSError:
            result = False
        return result

    def _merged_acronyms(self):
        """Return a single dictionary that merges acronyms from the search list."""
        # Start with an empty dict.
        result = {}
        # Parse files in reverse order since python dict.update() is "last one wins".
        for filename in reversed(self.search_list):
            result.update(self.acronyms_from(filename))
        return result

    def _search_list(self, prefer_files=None, exclude_files=None):
        """Return an ordered list of files in which to look for acronyms."""
        logging.debug("build search list in: %s", os.getcwd())
        if exclude_files:
            logging.warning("exclude files: %s", repr(exclude_files))
        else:
            exclude_files = []
        if prefer_files:
            logging.warning("prefer files: %s", repr(prefer_files))
        else:
            prefer_files = []

        # Fail early if user requests an impossible combination.
        if any(item in exclude_files for item in prefer_files):
            logging.error("One or more files is both excluded and preferred")
            sys.exit(1)

        # Find all yaml files.
        all_yamls = glob.glob("**/*", recursive=True)
        all_yamls = [f for f in all_yamls if self.is_yaml(f)]
        all_yamls.sort()
        logging.debug("All YAMLs: %s", repr(all_yamls))

        # Find YAML files with basename "global" in any subdirectory.
        # Ensure ROOT_GLOBAL comes first in the list of globals.
        global_yamls = [f for f in all_yamls if self.base_filename(f).lower() == "global"]
        if ROOT_GLOBAL in global_yamls:
            global_yamls = [g for g in global_yamls if g != ROOT_GLOBAL]
            global_yamls.insert(0, ROOT_GLOBAL)
        logging.debug("Global YAMLs: %s", repr(global_yamls))

        # Get the list of "other" yamls.
        other_files = all_yamls
        other_files = [f for f in other_files if f not in global_yamls]   # subtract globals
        other_files = [f for f in other_files if f not in prefer_files]   # subtract preferred files
        other_files = [f for f in other_files if f not in exclude_files]  # subtract excluded files
        logging.debug("Other YAMLs: %s", repr(other_files))

        # Construct search list.
        result = global_yamls        # Globals come first.
        result.extend(prefer_files)  # Preferred files come next.
        result.extend(other_files)   # Other files come last.
        logging.debug("Search list: %s", repr(result))
        return result



# Copyright (C) 2020 jumanjiman (Paul Morgan) <jumanjiman@gmail.com>
#
# This file is part of foal.
#
# foal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# foal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with foal.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""CLI utility to lookup acronyms from merged YAML files."""
import argparse
import logging
import os
import sys
from textwrap import dedent

from ruamel.yaml import YAML

from .acronyms import Acronyms
from .version import URL
from .version import VERSION

DEFAULT_DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

DESCRIPTION = dedent(
    """\
    Merge acronyms from YAML files in directory tree
    """,
)

EPILOG = dedent(
    """\
    search order:
      If an acronym is defined in multiple files, the first definition wins.
      The order is determined by the search list as follows:
        global.yaml in current directory
      + global.yaml in subdirectories
      + preferred files in order provided (-p)
      + all other YAMLs
      - excluded files (-x)
      ----------------------------------------------
      = search list

    environment variables:
      DEBUG=anything      Show verbose program operation
    """,
)


class OutputFormatter(YAML):
    """Trivial subclass of ruamel.yaml."""

    def __init__(self, **kwargs):
        """Create a YAML() instance from ruamel.yaml with settings as an output formatter."""
        super().__init__()
        self.indent(
            mapping=kwargs.get("mapping", 4),
            sequence=kwargs.get("sequence", 6),
            offset=kwargs.get("offset", 4),
        )
        self.explicit_start = kwargs.get("explicit_start", True)
        self.explicit_end = kwargs.get("explicit_end", False)
        self.width = kwargs.get("width", None)
        self.preserve_quotes = kwargs.get("preserve_quotes", False)
        self.top_level_colon_align = kwargs.get("colons", False)


class Cli:
    """Provide a command-line-interface for foal."""

    def __init__(self):
        """Return a new instance of the CLI."""
        self.build_parser()
        logging.basicConfig(
            format="[%(levelname)s] %(message)s",
            level=logging.DEBUG if "DEBUG" in os.environ else logging.WARNING,
        )

    def main(self):
        """Run the CLI."""
        args = self.parser.parse_args()

        try:
            original_dir = os.getcwd()
            if args.directory == DEFAULT_DATA_DIR:
                logging.warning("Using built-in data at %s", DEFAULT_DATA_DIR)
            os.chdir(args.directory)
            output = {}
            output["acronyms"] = Acronyms(
                args.prefer_files,
                args.exclude_files,
            ).get(args.acronyms, args.ignorecase)
            os.chdir(original_dir)
        except OSError as err:
            logging.error("%s: %s", err.strerror, err.filename)
            sys.exit(1)

        outfile = args.outfile
        try:
            outfile = open(outfile, "w")
        except OSError as err:
            logging.error("%s: %s", err.strerror, err.filename)
            sys.exit(1)
        except TypeError:
            outfile = sys.stdout

        OutputFormatter().dump(output, outfile)
        outfile.close()  # no effect on stdout

    def build_parser(self):
        """Private method to return the CLI argument parser."""
        parser = argparse.ArgumentParser(
            description=DESCRIPTION,
            epilog=EPILOG,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=f"foal {VERSION} {URL}",
        )
        parser.add_argument(
            "-a",
            metavar="ACRONYM",
            dest="acronyms",
            default=None,
            action="append",
            help="acronyms to find, can be used multiple times (default: all unique acronyms)",
        )
        parser.add_argument(
            "-i",
            dest="ignorecase",
            action="store_true",
            help="ignore case of acronyms",
        )
        parser.add_argument(
            "-o",
            metavar="FILENAME",
            dest="outfile",
            default=sys.stdout,
            help="output to FILENAME (default: STDOUT)",
        )
        parser.add_argument(
            "-p",
            metavar="FILENAME",
            dest="prefer_files",
            default=[],
            action="append",
            help="prefer FILENAME before other YAML files, can be used multiple times",
        )
        parser.add_argument(
            "-x",
            metavar="FILENAME",
            dest="exclude_files",
            default=None,
            action="append",
            help="exclude FILENAME, can be used multiple times  (cannot exclude global.yaml)",
        )
        parser.add_argument(
            metavar="DIRECTORY",
            dest="directory",
            nargs="?",
            default=DEFAULT_DATA_DIR,
            help=f"data directory (default: {DEFAULT_DATA_DIR})",
        )
        self.parser = parser



# setup.py
#
# See:
# https://packaging.python.org/guides/distributing-packages-using-setuptools/
# https://github.com/pypa/sampleproject
#
# Copyright (C) 2020 jumanjiman (Paul Morgan) <jumanjiman@gmail.com>
#
# This file is part of foal.
#
# foal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# foal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with foal.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""FOAL (From Org Acronyms Lookup) is a tool to lookup acronyms from a set of YAML files."""
from setuptools import find_packages
from setuptools import setup

from foal.version import URL
from foal.version import VERSION

# Get the long description from the README.
with open("README.md") as f:
    README_CONTENT = f.read()

setup(
    name="foal",
    version=VERSION,
    description="Merge and lookup acronyms from a directory tree of definitions.",
    long_description=README_CONTENT,
    long_description_content_type='text/markdown',
    license="GPLv3+",
    author="Paul Morgan",
    author_email="jumanjiman@gmail.com",
    keywords="acronyms lookup",
    url=URL,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "Intended Audience :: End Users/Desktop",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
    ],
    python_requires=">=3.5, <4",
    install_requires=[
        "identify>=1.4.18, <2",
        "ruamel.yaml>=0.16.10, <1",
    ],
    packages=find_packages(),
    package_data={
        "foal": [
            "data/*.yaml",
            "schema/*.yaml",
        ],
    },
    entry_points={
        "console_scripts": [
            "foal=foal.__main__:main",
        ],
    },
)
